openDevice=function(type,name,...) {
	switch(pmatch(type,c("tiff","eps")), {
		tiff(paste(name,".tif",sep=""),units="in",res=300,comp="lzw",...)
	},{
		postscript(paste(name,".eps",sep=""),paper="special",horizontal=F,...)
	})
}

get.interval.gamma=function(uncertainty,fac,num=100,maxInterval=100) {
	sapply(1:maxInterval,function(x){as.integer(sort(round(rgamma(num,shape=x^2/(uncertainty*(x^2*fac/10000+(1-fac))),scale=(uncertainty/x)*(x^2*fac/10000+(1-fac))))))})
}

best.layout=function(ngraphs,nc) {
	if(missing("nc")) nc=ifelse(ngraphs<3,ngraphs,3)
    nr=ceiling(ngraphs/nc)
    return(matrix(c(1:ngraphs,rep(0,nc*nr-ngraphs)),nc=nc,nr=nr,byr=T))
}

get.density=function(par) {
	ngr=get.n.groups(par)
	if(is.null(dim(par))) par=matrix(par,nr=1)
	par[,1:5]=floor(par[,1:5])
	vec=array(1:otherpars$maxInterval,dim=c(otherpars$maxInterval,dim(par)[1],2))
	vec[,,2]=0
	
	for(sol in 1:dim(par)[1]) {
		for(gr in 1:ngr[sol])
			vec[par[sol,gr],sol,2]=vec[par[sol,gr],sol,2]+par[sol,gr+5]
	}
	return(vec)
}

process.parameters=function(pars) {
	if(mode(pars)=="list") pars=pars$par
	if(dim(pars)[1]==0) return(pars)
	pars=cbind(pars,"% owners gr. 5"=NA)
	cs=apply(pars,1,function(a) {cumsum(a[6:9])})
	ngr=apply(cs<=1,2,sum)+1
	for(i in 1:dim(pars)[1]) {
		if(ngr[i]<5) {
			pars[i,(ngr[i]+1):5]=NA
			pars[i,(ngr[i]+5)]=1-ifelse(ngr[i]==1,0,sum(pars[i,6:(ngr[i]+4)]))
			pars[i,(ngr[i]+6):10]=NA
		} else pars[i,(ngr[i]+5)]=1-sum(pars[i,6:(ngr[i]+4)])
	}
	return(pars)
}

get.n.groups=function(pars) {	# gets the nr. of groups in the given decision variable set
	if(is.null(dim(pars)))
		return(sum(!is.na(pars[1:5])))
	else
		return(apply(!is.na(pars[,1:5,drop=F]),1,sum))
}

process.objectives=function(outlist) {
	inter=c(0,10,20,30,60,100)
	objectives=rep(list(array(dim=c(otherpars$n.years,NOBJ,otherpars$n.landscape.repetitions*otherpars$n.repeats),dimnames=list(NULL,otherpars$partial.obj.names,NULL))),length(outlist))
	if(is.null(outlist[[1]][[1]]$TSD)) notsd=T else notsd=F
	if(!notsd) TSD=rep(list(array(0,dim=c(otherpars$n.years,length(inter)-1,otherpars$n.landscape.repetitions*otherpars$n.repeats))),length(outlist)) else TSD=NULL

	for(i in 1:length(outlist)) {
		for(y in 1:otherpars$n.years) {
			for(l in 1:length(outlist[[i]])) {
				objectives[[i]][y,,l]=outlist[[i]][[l]]$Objectives[y,]
				
				if(!notsd) {
					b=table(findInterval(outlist[[i]][[l]]$TSD[,,y],inter,all.in=T))
					parc=b[as.character(1:(length(inter)-1))]
					parc[is.na(parc)]=0
					TSD[[i]][y,,l]=parc
				}
			}
		}
	
		objectives[[i]][,1,]=objectives[[i]][,1,]/otherpars$total.area*100
		objectives[[i]][,2,]=objectives[[i]][,2,]/otherpars$total.area
		objectives[[i]][,3,]=-objectives[[i]][,3,]
		objectives[[i]][,4,]=-objectives[[i]][,4,]
		for(j in 5:NOBJ) {	# remaining objectives are age class objectives, that can be changed
			objectives[[i]][,j,]=(objectives[[i]][,j,]/otherpars$total.area)*100
		}
	}
	
    somaglob=sum(TSD[[1]][1,,1]);
    TSD=lapply(TSD,function(x) {x/somaglob})
	return(list(obj=objectives,tsd=TSD))
}

format.pars=function(pars) {
    # formats optimization parameters for printing
    if(length(pars) %% 2 ==0) {
    	ng=sum(!is.na(pars[1:5]))
    	m.int=pars[1:ng]
    	prop=pars[6:(5+ng)]
    } else {
		ng=(length(pars)+1)/2
		
		ng1=sum(!cumsum(pars[(ng+1):(length(pars))])>=1)+1
		pars=pars[c(1:ng1,(ng+1):(ng1+ng-1))]
		ng=ng1
		m.int=pars[1:ng]
		
		if(any(is.na(m.int))) {
			ng=sum(!is.na(m.int))
			m.int=pars[1:ng]
			if(ng>1) {
				prop=pars[6:(6+ng-2)]
				prop=c(prop,1-sum(prop))
			} else prop=1;
		} else {
			if(ng>1) {
				prop=pars[(ng+1):length(pars)]
				prop=c(prop,1-sum(prop))
			} else prop=1;
		}
	}
	return(paste(round(prop*100),"%: ",ifelse(m.int>otherpars$maxInterval-1,"--",paste(floor(m.int),"yr",sep="")),sep="",collapse=" | "))
}

prepare.otherpars.for=function(otherpars,npars) {
	n.groups=(npars+1)/2
	if(n.groups==1) {
		parnames="Management interval"
	} else {
		parnames=c(paste("Man. interval gr.",1:n.groups),paste("% owners gr.",1:(n.groups-1)))
	}

	otherpars$npars=as.integer(n.groups*2-1)	# number of optimization parameters
	otherpars$parnames=parnames
	otherpars$lowerbound=c(rep(otherpars$minInterval,n.groups),rep(0.01,n.groups-1))	# lower and upper bounds where the parameters are allowed to vary
	otherpars$upperbound=c(rep(otherpars$maxInterval,n.groups),rep(1.01,n.groups-1))	# this is larger than 1 to allow for solution simplification to just one group
	return(otherpars)
}


processTasks=function(sch,pause=T) {
	results=list()
	for(i in 1:length(sch)) {
		if(sch[[i]]$task=="optim") {
		    cat("Using",otherpars$n.threads,"processor cores, repeating each solution",otherpars$n.repeats,"times in each of",otherpars$n.landscape.repetitions,"random landscapes.\n")
		    cat("Optimization will be run with a population of",otherpars$pop,"during",max(otherpars$gener),"generations. This may take a few days.\n")
			if(any(!file.exists(sch[[i]]$landscape.files))) stop("Some landscape files not found:\n",paste(sch[[i]]$landscape.files[!file.exists(sch[[i]]$landscape.files)],collapse=", "),"\nYou must create the needed landscape files prior to run this.")
			
			tasks=expand.grid(
				"landscape file"=sch[[i]]$landscape.files
				,uncertainty=sch[[i]]$uncertainties
				,incr.uncert=sch[[i]]$increasing.uncertainty
				,ngroups=sch[[i]]$n.groups
			)
			
			cat("The following optimizations will be performed:\n");
			print(tasks)
			if(pause) {
				cat("\nENTER to start.")
				readline()
			}
			for(j in 1:dim(tasks)[1]) {
				cat("******************************************NEW TASK**********************************\n")
				print(tasks[j,])
				
				# prepare optimization parameters
				otherpars$landscape.file=as.character(tasks[j,"landscape file"])
				n.groups=tasks[j,"ngroups"]
				otherpars$npars=as.integer((n.groups-1)*2+1)
				otherpars=prepare.otherpars.for(otherpars,otherpars$npars)
				otherpars$rgama=get.interval.gamma(tasks[j,"uncertainty"],tasks[j,"incr.uncert"])

				# draw uncertainty distributions (boxplots)
				openDevice(output,name=paste("uncert_",tasks[j,"uncertainty"],"_",tasks[j,"incr.uncert"],sep=""),wid=7,hei=7)
				plotUncertainty(otherpars$rgama)
				dev.off()
				
				# start optimization
				original.gener=otherpars$gener
				otherpars$gener=as.integer(otherpars$gener)
				cat("Generations:",otherpars$gener,"\n")
				tmpres=nsga2(objective.fn,otherpars$npars,NOBJ
					,generations=otherpars$gener,popsize=otherpars$pop
					,lower.bounds=otherpars$lowerbound
					,upper.bounds=otherpars$upperbound
					,objfuncpars=otherpars
				)
				otherpars$gener=original.gener
				if(length(otherpars$gener)==1) {
					tmpres=list(tmpres)
				}
				for(k in 1:length(tmpres)) {
					colnames(tmpres[[k]]$value)=otherpars$obj.names
					colnames(tmpres[[k]]$par)=otherpars$parnames
				}

				results=c(results,list(tmpres))
				names(results)[length(results)]=paste(otherpars$landscape.file,"_",tasks[j,"uncertainty"],sep="")
				save(results,file=paste("optim_partial_results_",paste(LETTERS[sample(1:length(LETTERS),6)],collapse=""),".rdata",sep=""))
			}
			out=results
		}
		
		if(sch[[i]]$task=="examine") {
			uncertainties=sch[[i]]$uncertainties
			solutions=sch[[i]]$solutions
			
			otherpars$landscape.file=sch[[i]]$landscape.files
			otherpars$trace=as.integer(0:(otherpars$n.years-1))
			otherpars$savelandscapes=T

			outlist=list()
			counter=1
		
			cat("Simulating solutions...\n")
			flush.console()
			for(j in 1:length(uncertainties)) {
				uncertainty=uncertainties[j]
				fac=sch[[i]]$increasing.uncertainty[j]
				otherpars$rgama=get.interval.gamma(uncertainty,fac)
						
				for(solution in solutions) {
					otherpars=prepare.otherpars.for(otherpars,length(solution))
					outlist[[counter]]=simulateOneSolution(solution,otherpars)
					names(outlist)[counter]=paste(paste(solution,collapse=" "),uncertainty)
					counter=counter+1
				}
			}
			
			results=outlist
			
			cat("Processing yearly objectives and landscapes...\n")
			flush.console()
		# process yearly objectives			
			inter=c(0,10,20,30,60,100)
			objectives=rep(list(array(dim=c(otherpars$n.years,NOBJ,otherpars$n.landscape.repetitions*otherpars$n.repeats),dimnames=list(NULL,otherpars$partial.obj.names,NULL))),length(outlist))
			if(is.null(outlist[[1]][[1]]$TSD)) notsd=T else notsd=F
			if(!notsd) TSD=rep(list(array(0,dim=c(otherpars$n.years,length(inter)-1,otherpars$n.landscape.repetitions*otherpars$n.repeats))),length(outlist)) else TSD=NULL

			curtask=i
			for(i in 1:length(outlist)) {
				for(y in 1:otherpars$n.years) {
					for(l in 1:length(outlist[[i]])) {
						objectives[[i]][y,,l]=outlist[[i]][[l]]$Objectives[y,]
				
						if(!notsd) {
							b=table(findInterval(outlist[[i]][[l]]$TSD[,,y],inter,all.in=T))
							parc=b[as.character(1:(length(inter)-1))]
							parc[is.na(parc)]=0
							TSD[[i]][y,,l]=parc
						}
					}
					if(y %% 10==0) cat(y) else cat(".")
					flush.console()
				}
				cat("\n")
				objectives[[i]][,1,]=objectives[[i]][,1,]/otherpars$total.area*100
				objectives[[i]][,2,]=objectives[[i]][,2,]/otherpars$total.area
				objectives[[i]][,3,]=-objectives[[i]][,3,]
				objectives[[i]][,4,]=-objectives[[i]][,4,]
				for(j in 5:NOBJ) {	# remaining objectives are age class objectives, that can be changed
					objectives[[i]][,j,]=(objectives[[i]][,j,]/otherpars$total.area)*100
				}
			}
			i=curtask
			somaglob=sum(TSD[[1]][1,,1]);
			TSD=lapply(TSD,function(x) {x/somaglob})
			objectives=list(obj=objectives,tsd=TSD)
		#####			
			names(objectives$obj)=names(outlist)
			names(objectives$tsd)=names(outlist)
		# plot objective values along time
			plotObjectives(objectives$obj,obj.names=otherpars$partial.obj.names,legend=sch[[i]]$legend.func)
			for(i in 1:length(objectives$tsd)) {
				plotAreas(objectives$tsd[i],file=paste("areas_sol",i,sep=""))
			}
			out=objectives
		}
	}
	return(out)
}

# set constraints used for setting objective preferences, based on percentiles
setLimits=function(results) {
	limits=list(
		"No restrictions"=NULL
		,"Biodiversity priority"=list(NA,NA,NA,NA,c(9.9,11),c(9.9,11))
		,"Fire risk priority"={
			percentis=list(numeric(0),numeric(0),numeric(0))
			percentis[[1]]=c(percentis[[1]],results$value[,2])
			percentis[[2]]=c(percentis[[2]],results$value[,3]*-1)
			percentis[[3]]=c(percentis[[3]],results$value[,4]*-1)
			percentis=mapply(quantile,percentis,c(0.25,0.75,0.75))
			list(NA,c(-1000,percentis[1]),c(percentis[2],4000000),c(percentis[3],100),NA,NA);
		}
		,"Cost efficiency - 5% cleared area"=list(5+c(-0.375,0.375),NA,NA,NA,NA,NA)
		,"Cost efficiency - 10% cleared area"=list(10+c(-0.375,0.375),NA,NA,NA,NA,NA)	
		,"Compromise solutions"={
			percentis=list(numeric(0),numeric(0),numeric(0))
			for(i in 1:length(results)) {
				percentis[[1]]=c(percentis[[1]],results$value[,2])
				percentis[[2]]=c(percentis[[2]],results$value[,3]*-1)
				percentis[[3]]=c(percentis[[3]],results$value[,4]*-1)
			}
			percentis=mapply(quantile,percentis,c(0.6,0.4,0.4))	# e.g. the best 10% would be 0.1 0.9 0.9
			list(c(-100,20),c(-1000,percentis[1]),c(percentis[2],4000000),c(percentis[3],100),c(9,11),c(9,11))
		}
	)
	return(limits)
}

compute.distance=function(r,norm="euclidean",weights=c(6,2,2,2,3,3)) {
	ran=apply(r$value,2,range)

	for(i in 1:dim(r$value)[2]) {	# for each objective, range standardize, giving weights so that cost, risk and biodiversity contribute 1/3 each
		if(ran[2,i]-ran[1,i]==0) r$value[,i]=0 else {
			r$value[,i]=( (r$value[,i]-ran[1,i])/(ran[2,i]-ran[1,i]) ) * weights[i]
		}
	}
	normv=apply(r$value/(sum(weights)/6),1,function(x){
		switch(pmatch(norm,c("1-norm","euclidean","infinite"))
			,sum(abs(x))
			,sqrt(sum(x^2))
			,max(x))
	})
    return(normv)
}

subset=function(r,limits) {
	cond=rep(T,dim(r$value)[1])
	for(i in 1:length(limits)) {
		if(length(limits[[i]])==2) {
			cond=cond & r$value[,i]>=limits[[i]][1] & r$value[,i]<=limits[[i]][2]
		}
	}
	highlight=which(cond)
	r$par=r$par[highlight,,drop=F]
	r$value=r$value[highlight,,drop=F]
    return(list(which.solutions=highlight,subsetted.results=r))
}

get.best.overall=function(norm,whichbest=1:10,type=c("overall","each")[1]) {
	if(length(norm)==0) return(NULL)
	if(is.na(whichbest[1])) whichbest=1:length(norm)
	return(order(norm)[whichbest])
}


# create a new hexagonal landscape grid
newHexgrid=function(xlim,ylim,r,a,t,buffer=0) {
    # Specify one of:
    # 	r: desired radius (of the inscribed circle)
    # 	t: desired side (of the hexagon)
    # 	a: desired hexagon area
    
    # buffer: expand grid by "buffer" hexagons in all sides
    
    if(!missing(t)) {
    	a=(3^(3/2)*t^2)/2
		r=(sqrt(3)/2)*t
    } else if(!missing(a)) {
		r=sqrt(a)/(sqrt(2)*3^0.25)
		t=(2*r)/sqrt(3)
    } else {
		a=2*sqrt(3)*r^2
		t=(sqrt(2)*sqrt(a))/3^(3/4)
    }
    cat("Area:",a,"\nRadius:",r,"\nSide:",t,"\n")
    
    if(buffer>0) {
		x=t+sqrt(t^2-r^2)
		ylim[1]=ylim[1]-x*buffer-1	# because of rounding
		ylim[2]=ylim[2]+x*buffer+1
		xlim[1]=xlim[1]-r*2*buffer-1
		xlim[2]=xlim[2]+r*2*buffer+1
    }
    
    d=r*2
    vert=t*1.5
    nc=round((xlim[2]-xlim[1])/d)+1
    nr=ceiling( (ylim[2]-ylim[1]-t)/ (1.5*t) )+1

    out=list(id=matrix(seq.int(1,nc*nr),nc=nc,nr=nr,byr=T),data=list())
    
    out$edge=matrix(as.integer(0),nc=2*nc-1,nr=2*nr-1)		# contiguity relationships
    out$edge[c(T,F),c(F,T)]=-1L		# code for horizontal contiguity
    out$edge[c(F,T),c(T,F)]=-2L		# code for vertical contiguity
    out$edge[c(F,T,F,F),c(F,T)]=-3L	# code for SE-NW contiguity
    out$edge[c(F,F,F,T),c(F,T)]=-4L	# code for SW-NE contiguity
    
    attr(out,"xlim")=xlim
    attr(out,"ylim")=ylim
    attr(out,"size")=c(nr,nc)
    names(attr(out,"size"))=NULL
    attr(out,"metrics")=c(radius=r,side=t,area=a)
    class(out)="hexgrid"
    return(out)
}

