#include "simulator.h"
#ifdef NONAS
#pragma message "Assuming there are no NAs in the landscapes!"
#endif
#ifdef KEEPSTRATEGYGROUPS
#pragma message "Forcing algorithm to maintain the given number of groups!"
#define ABORT_SOLUTION(m) do {for(i=0;i<NOBJ;i++) *out[i]=1000000;Rprintf(m);return;} while(0)
#endif

static void permute(int *permutation,int n);
int threadcounter;

inline void swap(double *a,double *b) {
	double c=*a;
	*a=*b;
	*b=c;
}

void *generateLandscape(void *thrdata) {
	THREADDATA *tdt=(THREADDATA*)thrdata;
	register int i,j,k,m;
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(VECTOR_ELT(oriland,1),1)));	// landscape dimensions
	int	*dimedge=INTEGER_POINTER(GET_DIM(VECTOR_ELT(oriland,2)));
	int lengedge=dimedge[0]*dimedge[1],dimedge2=2*dimedge[0];
	int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));					// nr. of cells in the landscape (count of ownerID field)
	OWNERCELLS *tmpown;

	// generate new landscape
	randLandscape(oriland,landscapegamma,tdt->workinglandscape->edge);

	for(i=0,k=0;i<dimedge[0];i+=2,k++) {		// copy owner IDs into our array
		for(j=0,m=k;j<lengedge;j+=dimedge2,m+=dim[0]) tdt->workinglandscape->ownerIDs[m]=-tdt->workinglandscape->edge[j+i];
	}
	for(i=0;i<ncells;i++) tdt->workinglandscape->tsd[i]=unifTSD[tdt->workinglandscape->ownerIDs[i]];
/*		
Rprintf("\n****************** NEW LANDSCAPE ************************\n");
for(x=0,i=0;x<dim[1];x++) {
	for(y=0;y<dim[0];y++,i++) {					
		if(x<dim[1]/8 && y<dim[0]/1) Rprintf("%c",64+(int)(((float)(workinglandscape.tsd[i])/(float)100)*48));
	}
	if(x<dim[1]/8) Rprintf("\n");
}
Rprintf("***************************************************\n");*/
		
	// build and index of the cells owned by each owner, and nr. of owners of this landscape
	tdt->workinglandscape->nowners=0;
	for(i=0;i<ncells;i++) {
		if(tdt->workinglandscape->ownerIDs[i]>tdt->workinglandscape->nowners) tdt->workinglandscape->nowners=tdt->workinglandscape->ownerIDs[i];
#ifdef ERRORCHECK
		if(tdt->workinglandscape->nowners>=MAXNOWNERS) error("you have to recompile and increase MAXNOWNERS");
#endif
		tmpown=&tdt->workinglandscape->ownercells[tdt->workinglandscape->ownerIDs[i]-1];
#ifdef ERRORCHECK
		if(tmpown->ncells>=MAXOWNEDCELLS) error("you have to recompile and increase MAXOWNEDCELLS");
#endif
		tmpown->ownedcells[tmpown->ncells]=i;
		tmpown->ncells++;
	}
//Rprintf("# owners: %d\n",nowners);		
	for(i=0;i<tdt->workinglandscape->nowners;i++) tdt->workinglandscape->ownerindexes[i]=i;

	tdt->ready=1;
	tdt->returned=1;
	return 0;	
}

SEXP makeLandscapes(SEXP nlands,SEXP modelland,SEXP lgamma,SEXP maxano,SEXP _nthreads,SEXP filename) {
	register int i,j,k,x,y;
	int curlandscape,tmp,soma;
	int nland=INTEGER_POINTER(nlands)[0];
	int max_ano=INTEGER_POINTER(maxano)[0];
	int nthreads=INTEGER_POINTER(_nthreads)[0];
	int jobcounter;
	pthread_t threads[MAXCORETHREADS];
	pthread_attr_t attr;
	
	oriland=modelland;
	landscapegamma=lgamma;
	memset(td,0,sizeof(THREADDATA)*nthreads);

	int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));					// nr. of cells in the landscape (count of ownerID field)
	int	*dimedge=INTEGER_POINTER(GET_DIM(VECTOR_ELT(oriland,2)));
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(VECTOR_ELT(oriland,1),1)));	// landscape dimensions
	workinglandscapes=malloc(sizeof(WORKINGLANDSCAPE)*nland);
	for(i=0;i<nland;i++) {
		workinglandscapes[i].tsd=(int*)malloc(sizeof(int)*ncells);
		workinglandscapes[i].ownerIDs=(int*)malloc(sizeof(int)*ncells);
		workinglandscapes[i].edge=malloc(sizeof(int)*dimedge[0]*dimedge[1]);
	}

	for(i=0;i<MAXNOWNERS;i++) unifTSD[i]=(int)((float)rand()/RAND_MAX*max_ano);

	for(i=0;i<nthreads;i++) {
		td[i].ready=1;		// thread ready for working?
		td[i].returned=0;	// thread returned value?
	}
		
// generate N random landscapes (in separate threads)
	curlandscape=0;
	threadcounter=0;jobcounter=0;
	do {		// landscape loop: generate new landscape each time
		for(k=0;k<nthreads;k++) {if(td[k].ready) break;}
		if(k<nthreads) {
			if(td[k].returned) {	// save results from previous
				jobcounter++;
				threadcounter--;
				Rprintf("Landscape %d: # owners %d\n",curlandscape,td[k].workinglandscape->nowners);
				for(x=0,i=0;x<dim[1];x++) {
					for(y=0;y<dim[0];y++,i++) {					
						if(x<dim[1]/8 && y<dim[0]/2) Rprintf("%c",64+(int)(((float)(td[k].workinglandscape->tsd[i])/(float)100)*48));
					}
					if(x<dim[1]/8) Rprintf("\n");
				}
				Rprintf("***************************************************\n");
				td[k].ready=(curlandscape>=nland ? 0 : 1);		// signal to kill thread: ready=0, returned=0
				td[k].returned=0;
			}
		
			if(curlandscape<nland) {
				td[k].ready=0;
				td[k].returned=1;
				td[k].workinglandscape=&workinglandscapes[curlandscape];
				
				curlandscape++;
				
				pthread_attr_init(&attr);
				pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
				tmp=pthread_create(&threads[k],&attr,generateLandscape,(void*)&td[k]);
				if(tmp!=0) {td[k].ready=1; threadcounter--; Rprintf("ERROR: %d\n",tmp);}
				pthread_attr_destroy(&attr);
				threadcounter++;
			} else {
				if(!threadcounter && jobcounter==nland) break;
			}
		} else {
			if(!threadcounter && jobcounter==nland) break;
		}
	} while(1);

	const char *fname=CHARACTER_VALUE(filename);
	Rprintf("Writing landscapes to file %s\n",fname);
	FILE *f;
	size_t wb;
	f=fopen(fname,"w");
	wb=fwrite(&nland,sizeof(int),1,f);
	wb=fwrite(workinglandscapes,sizeof(WORKINGLANDSCAPE),nland,f);
	for(i=0;i<nland;i++) {
		wb=fwrite(workinglandscapes[i].ownerIDs,sizeof(int)*ncells,1,f);
		wb=fwrite(workinglandscapes[i].tsd,sizeof(int)*ncells,1,f);
		wb=fwrite(workinglandscapes[i].edge,sizeof(int)*dimedge[0]*dimedge[1],1,f);
		Rprintf("Landscape %d, nowners: %d\n",i,workinglandscapes[i].nowners,wb);
	}
	fclose(f);
	
	for(i=0;i<nland;i++) {
		free(workinglandscapes[i].tsd);
		free(workinglandscapes[i].ownerIDs);
		free(workinglandscapes[i].edge);
	}

	free(workinglandscapes);
	
	return(R_NilValue);
}

SEXP simulateOneSolution(SEXP par,SEXP objfuncpars) {
	SEXP ret;
	prepare(objfuncpars);
	PROTECT(ret=NEW_LIST(nlandscapes*nrepeats));
	int i;
	int npars=INTEGER_POINTER(getListElement(objfuncpars,"npars"))[0];
	int NOBJ=4+LENGTH(minareaTSD);
	double *pars=NUMERIC_POINTER(par);
	double *obj=malloc(sizeof(double)*NOBJ);
	double *ppars[15],*pobj[15];
	for(i=0;i<npars;i++) ppars[i]=&pars[i];
	for(i=0;i<NOBJ;i++) pobj[i]=&obj[i];
	objectiveFunction(ppars,pobj,objfuncpars,ret);
	free(obj);
	cleanUp();
	UNPROTECT(1);
	return(ret);
}

void objectiveFunction(double **oripar,double **out,SEXP objfuncpars,SEXP tracelist) {
	register int i,j,k;
	int soma,tmp,start,end,curlandscape,currep,jobcounter,protcount=0;
	double somaprop,*pobj,*pminareaTSD=NUMERIC_POINTER(minareaTSD);
	int npars=INTEGER_POINTER(getListElement(objfuncpars,"npars"))[0];
	int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));					// nr. of cells in the landscape (count of ownerID field)
	int NOBJ=4+LENGTH(minareaTSD);
	int nthreads=INTEGER_POINTER(getListElement(objfuncpars,"n.threads"))[0];
	int manintervals[15]={0};
	int maxInterval=INTEGER_POINTER(getListElement(objfuncpars,"maxInterval"))[0];
// for displaying progress	
	int ngener=INTEGER_POINTER(getListElement(objfuncpars,"gener"))[ LENGTH(getListElement(objfuncpars,"gener"))-1 ];
	int pop=INTEGER_POINTER(getListElement(objfuncpars,"pop"))[0];
// for tracing
	int *intervalostsd=INTEGER_POINTER(intervals);
	short nintervalostsd=GET_LENGTH(intervals)-1;
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(VECTOR_ELT(oriland,1),1)));	// landscape dimensions
	char tmpbuf[120];
	int lentrace=0,nmanint,npropor,effect_nmanint;
	SEXP trcarga,trtsd,trobj,obj,names,dnames,colnames;
	double par[MAXPARS];
	for(i=0;i<npars;i++) par[i]=*oripar[i];		// make local copy of parameters
	for(i=npars;i<MAXPARS;i++) par[i]=1;		// make local copy of parameters

	pthread_t threads[MAXCORETHREADS];
	pthread_attr_t attr;

int x,y;	
	
#ifdef KEEPSTRATEGYGROUPS	// check if this is a viable solution in terms of maintaining the given number of groups
	effect_nmanint=nmanint;
// first, check proportions
	for(i=0,somaprop=0;i<npropor;i++) somaprop+=par[i+nmanint];		// par[i+nmanint] are the proportion of owners in each group
	if(somaprop>0.99) ABORT_SOLUTION("Aborted, impossible proportions.\n");	// this is not a viable solution because it would nearly reduce to (n.groups-1) groups, abort 
// now check management intervals - if there are duplicate intervals, abort, because it would reduce to (n.groups-1) groups
	int tmparr[100]={0};
	for(i=0;i<nmanint;i++) {
		if(tmparr[(int)floor(par[i])]) ABORT_SOLUTION("Aborted, duplicate management intervals.\n"); else tmparr[(int)floor(par[i])]=1;
	}
#else	// otherwise, adjust parameters accordingly: sort proportions decreasingly and change proportions to 0 when cumsum>100%
	nmanint=(npars+1)/2;
	npropor=(npars-1)/2;
	effect_nmanint=1;

// prepare proportions and count nr. of groups by the cumulative sum
	for(i=0,somaprop=0;i<npropor+1;i++) {
		if(somaprop+par[i+nmanint]>1) {
			par[i+nmanint]=1-somaprop;
			for(j=i+1;j<npropor+1;j++) {
				par[j+nmanint]=0;
				par[j]=0;
			}
			break;
		} else effect_nmanint++;
		somaprop+=par[i+nmanint];
	}
#endif

// prepare management intervals	
	for(i=0;i<NOBJ;i++) *out[i]=-100000000000;
	for(i=0;i<nmanint;i++) {
		if(par[i]>maxInterval-1) manintervals[i]=NA_INTEGER; else manintervals[i]=(int)floor(par[i]);
	}

// display parameters of this solution	
	Rprintf("%d%; %d groups; Mng.int.: ",(int)((float)progress/(((float)ngener+2)*pop)*100),effect_nmanint);
	for(i=0;i<nmanint;i++) {if(manintervals[i]==NA_INTEGER) Rprintf(" - "); else if(manintervals[i]==0) Rprintf("   "); else Rprintf("%2d ",manintervals[i]);}
	Rprintf(" Prop.: ");
	for(i=nmanint;i<npars+1;i++) {if(par[i]==0) Rprintf("     "); else Rprintf("%.2f ",par[i]);}
#ifdef VERBOSE
	Rprintf("\n");
#endif

// initialize threaddata structure
	memset(td,0,sizeof(THREADDATA)*nthreads);
	for(i=0;i<nthreads;i++) {
		td[i].ready=1;		// thread ready for working?
		td[i].returned=0;	// thread returned value?
	}
	
	if(landscapefile==R_NilValue) {		// no landscape file given, have to generate random landscapes on the fly
	// generate N random landscapes (in separate threads)
		curlandscape=0;
		do {		// landscape loop: generate new landscape each time
			for(k=0;k<nthreads;k++) {if(td[k].ready) break;}
			if(k<nthreads) {
				if(td[k].returned) {	// save results from previous
#ifdef VERBOSE			
					Rprintf("New landscape. Nowners: %d\n",td[k].workinglandscape->nowners);
#ifdef SHOWLANDSCAPES
					Rprintf("\n****************** NEW LANDSCAPE ************************\n");
					for(x=0,i=0;x<dim[1];x++) {
						for(y=0;y<dim[0];y++,i++) {					
							if(x<dim[1]/8 && y<dim[0]/1) Rprintf("%c",64+(int)(((float)(td[k].workinglandscape->tsd[i])/(float)100)*48));
						}
						if(x<dim[1]/8) Rprintf("\n");
					}
					Rprintf("***************************************************\n");
#endif
#else
					Rprintf(".");
#endif
					td[k].ready=(curlandscape>=nlandscapes ? 0 : 1);		// signal to kill thread: ready=0, returned=0
					td[k].returned=0;
	//	Rprintf("Thread %d ready: %d; returned: %d\n",k,td[k].ready,td[k].returned);
				}
		
				if(curlandscape<nlandscapes) {
					td[k].ready=0;
					td[k].returned=1;
					td[k].workinglandscape=&workinglandscapes[curlandscape];
				
					curlandscape++;
				
					pthread_attr_init(&attr);
					pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
					tmp=pthread_create(&threads[k],&attr,generateLandscape,(void*)&td[k]);
					if(tmp!=0) {td[k].ready=1; Rprintf("ERROR: %d\n",tmp);}
					pthread_attr_destroy(&attr);
				} else {
					for(j=0,soma=0;j<nthreads;j++) {	// check if all threads are free and exit if so
						if(!td[j].ready && !td[j].returned) soma++;
					}
					if(soma==nthreads) break;			// all threads done, exit repetition loop
				}
			}
		} while(1);
	
#ifdef VERBOSE
		for(i=0;i<nlandscapes;i++) {
			Rprintf("Landscape %d, nowners: %d\n",i,workinglandscapes[i].nowners);
		}
#endif
	} else loadLandscapesFromFile();	// landscape file given

/********************************** landscapes done (or read from file), go for solution evaluation *******************************/
	for(i=0;i<nthreads;i++) {
		td[i].pobj=(double*)malloc(sizeof(double)*NOBJ);
		td[i].manint=manintervals;
		td[i].tsd=(int*)malloc(sizeof(int)*ncells);
		td[i].ready=1;		// thread ready for working?
		td[i].returned=0;	// thread returned value?
		if(trace!=R_NilValue) {		// do we need to save values during simulations? allocate space then.
			PROTECT(td[i].tracelist=NEW_LIST(3));
			protcount+=1;
			PROTECT(names=NEW_CHARACTER(3));
			SET_STRING_ELT(names,0,mkChar("Biomass"));
			SET_STRING_ELT(names,1,mkChar("Objectives"));
			SET_STRING_ELT(names,2,mkChar("TSD"));

			PROTECT(dnames=NEW_LIST(2));
			PROTECT(colnames=NEW_CHARACTER(NOBJ));
			SET_STRING_ELT(colnames,0,mkChar("Total managed area (% from total)"));
			SET_STRING_ELT(colnames,1,mkChar("Mean biomass per hectare"));
			SET_STRING_ELT(colnames,2,mkChar("Mean edge contrast variance (-)"));
			SET_STRING_ELT(colnames,3,mkChar("Mean biomass skewness (-)"));
			for(j=0;j<nintervalostsd;j++) {
				sprintf(tmpbuf,"Area with %d<TSD<=%d",intervalostsd[j],intervalostsd[j+1]);
				SET_STRING_ELT(colnames,4+j,mkChar(tmpbuf));
			}
		
			SET_VECTOR_ELT(dnames,0,R_NilValue);
			SET_VECTOR_ELT(dnames,1,colnames);

			lentrace=LENGTH(trace);
			if(savelandscapes) {
				PROTECT(trcarga=alloc3DArray(INTSXP,dim[0],dim[1],lentrace));
				PROTECT(trtsd=alloc3DArray(INTSXP,dim[0],dim[1],lentrace));
			} else {
				PROTECT(trcarga=R_NilValue);
				PROTECT(trtsd=R_NilValue);
			}
			PROTECT(trobj=allocMatrix(REALSXP,lentrace,NOBJ));
			SET_DIMNAMES(trobj,dnames);
			SET_VECTOR_ELT(td[i].tracelist,0,trcarga);
			SET_VECTOR_ELT(td[i].tracelist,1,trobj);
			SET_VECTOR_ELT(td[i].tracelist,2,trtsd);
			SET_NAMES(td[i].tracelist,names);
			UNPROTECT(6);
		} else td[i].tracelist=0;
	}

	curlandscape=0;currep=0;threadcounter=0;jobcounter=0;
	do {							// repetition loop: evaluate solution n times in each landscape
		for(k=0;k<nthreads;k++) {
			if(td[k].ready) break;		// a thread is ready
		}
		if(k<nthreads) {
			if(td[k].returned) {	// one thread has returned, save results
				jobcounter++;
				threadcounter--;
				pobj=td[k].pobj;
				pobj[0]=pobj[0]/(totalarea*nyears);		// % managed area per year
				pobj[1]=pobj[1]/totalarea;				// max biomass per hectare
				pobj[2]=-pobj[2];						// mirror, since the algorithm minimizes
				pobj[3]=-pobj[3];
				for(i=4;i<NOBJ;i++) {	// if minimum attained area within each age class is higher than the required minimum of 10% (of total area), keep it constant for optimization purposes. Otherwise, the value is the (inverted) proportion of those 10% that are safeguarded. 
					if(pobj[i]/totalarea >= pminareaTSD[i-4]) pobj[i]=-1; else pobj[i]=-pobj[i]/(totalarea*pminareaTSD[i-4]);
				}
				if(trace!=R_NilValue) SET_VECTOR_ELT(tracelist,jobcounter-1,duplicate(td[k].tracelist));
#ifdef VERBOSE				
				Rprintf("Job #%d: landscape nowners %d: %.2f %.2f %.2f %.2f %.2f %.2f\n",jobcounter,td[k].workinglandscape->nowners,pobj[0],pobj[1],pobj[2],pobj[3],pobj[4],pobj[5]);				
#endif
/*for(x=0,i=0;x<dim[1];x++) {
for(y=0;y<dim[0];y++,i++) {					
	if(x<dim[1]/8) Rprintf("%c",64+(int)(((float)(td[k].tsd[i])/(float)100)*48));
}
if(x<dim[1]/8) Rprintf("\n");
}
Rprintf("***************************************************\n");*/
			
				for(i=0;i<NOBJ;i++) *out[i]=max(*out[i],pobj[i]);
				td[k].ready=(curlandscape>=nlandscapes ? 0 : 1);		// signal to kill thread: ready=0, returned=0
				td[k].returned=0;
			}
			if(curlandscape<nlandscapes) {			// there's work to do, create another thread
				td[k].ready=0;
				td[k].returned=1;
				td[k].workinglandscape=&workinglandscapes[curlandscape];
				if(npars==1) {
					for(i=0;i<td[k].workinglandscape->nowners;i++) td[k].ownerlist[i].group=0;
				} else {
					// this randomly assigns a group to each owner based on the given desired proportions in each group.
					permute(td[k].workinglandscape->ownerindexes,td[k].workinglandscape->nowners);
					start=0;
					for(i=0;i<npropor;i++) {
						end=lround(par[i+nmanint] * td[k].workinglandscape->nowners)+start-1;
						for(j=start;j<end;j++) td[k].ownerlist[td[k].workinglandscape->ownerindexes[j]].group=i;
						start=end+1;
					}
					for(j=start;j<td[k].workinglandscape->nowners;j++) td[k].ownerlist[td[k].workinglandscape->ownerindexes[j]].group=i;
					
//					for(i=0;i<td[k].workinglandscape->nowners;i++) Rprintf("%d ",td[k].ownerlist[i].group);
				}
				
#ifndef VERBOSE
				if(currep==0) Rprintf(".");
#else
//				Rprintf("Landscape %d, repetition %d, jobs done %d... ",curlandscape,currep,jobcounter);
#endif
//Rprintf("Landscape %d, repetition %d... ",curlandscape,currep);
				currep++;
				if(currep==nrepeats) {curlandscape++;currep=0;}

				memset(td[k].pobj,0,sizeof(double)*NOBJ);
				memcpy((void*)td[k].tsd,(void*)td[k].workinglandscape->tsd,sizeof(int)*ncells);		// copy landscape TSD from original (generated previously)
/*Rprintf("\n********************STARTING***********************\n");
for(x=0,i=0;x<dim[1];x++) {
for(y=0;y<dim[0];y++,i++) {					
	if(x<dim[1]/8) Rprintf("%c",64+(int)(((float)(td[k].tsd[i])/(float)100)*48));
}
if(x<dim[1]/8) Rprintf("\n");
}
Rprintf("***************************************************\n");*/
			
				pthread_attr_init(&attr);
				pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
				tmp=pthread_create(&threads[k],&attr,simulator,(void*)&td[k]);
				if(tmp!=0) {td[k].ready=1; threadcounter--; Rprintf("ERROR: %d\n",tmp);}
				pthread_attr_destroy(&attr);
				threadcounter++;
			} else {		// no more work to do, check if all threads are done
				if(!threadcounter && jobcounter==nlandscapes*nrepeats) break;
			}
		} else {
			if(!threadcounter && jobcounter==nlandscapes*nrepeats) break;
		}
	} while(1);		// END repetition loop
	
	if(protcount) UNPROTECT(protcount);
	
// cleanup owerships
	for(i=0;i<nlandscapes;i++) {
		for(j=0;j<workinglandscapes[i].nowners;j++) workinglandscapes[i].ownercells[j].ncells=0;
	}

	for(i=0;i<nthreads;i++) {
		free(td[i].pobj);
		free(td[i].tsd);
	}
	
// display progress	
	char buf[200];
	memset(buf,0,200);
	memset(buf,'\b',nlandscapes+(landscapefile==R_NilValue ? nlandscapes : 0));
	strcat(buf,"FINAL Obj: %.2f %.2f %.2f %.2f");
	Rprintf(buf,*out[0],*out[1],*out[2],*out[3]);
	for(i=4;i<NOBJ;i++) Rprintf(" %.2f",*out[i]);
	Rprintf(" (n=%d)\n",jobcounter);
//	Rprintf(buf,*out[0],*out[1],*out[2],*out[3],*out[4],*out[5],jobcounter);
//	R_CheckUserInterrupt();
	progress++;
}

void loadLandscapesFromFile() {
		register int i,j,k;
		int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));					// nr. of cells in the landscape (count of ownerID field)
		int	*dimedge=INTEGER_POINTER(GET_DIM(VECTOR_ELT(oriland,2)));
		FILE *lf=fopen(CHARACTER_VALUE(landscapefile),"r");
		int totallands;
		int *tmpt1,*tmpt2,*tmpt3;
		j=fread(&totallands,sizeof(int),1,lf);
		if(!j) error("Error #1 reading landscape file.");
		
#ifdef VERBOSE		
		Rprintf("%d landscapes found in file.\nReading landscapes #",totallands);
#endif
		for(k=0;k<nlandscapes;k++) {
			if(onelandscapeonly==R_NilValue)
				j=(int)((float)rand()/RAND_MAX*totallands);
			else {
				if(INTEGER_POINTER(onelandscapeonly)[0]==NA_INTEGER)
					j=(int)((float)rand()/RAND_MAX*totallands);
				else
					j=INTEGER_POINTER(onelandscapeonly)[0];
			}
			
#ifdef VERBOSE
			Rprintf("%3d ",j);
#endif
			tmpt1=workinglandscapes[k].ownerIDs;		// pointers were stored in file for comodity, but are meaningless. save malloc'ed pointers before overriding values
			tmpt2=workinglandscapes[k].tsd;
			tmpt3=workinglandscapes[k].edge;
			
			fseek(lf,sizeof(int)+sizeof(WORKINGLANDSCAPE)*j,SEEK_SET);
			i=fread(&workinglandscapes[k],sizeof(WORKINGLANDSCAPE),1,lf);
#ifdef VERBOSE			
			Rprintf("%d owners.\n",workinglandscapes[k].nowners);
#endif			
			workinglandscapes[k].ownerIDs=tmpt1;
			workinglandscapes[k].tsd=tmpt2;
			workinglandscapes[k].edge=tmpt3;

			fseek(lf,sizeof(int)+sizeof(WORKINGLANDSCAPE)*totallands+(sizeof(int)*ncells*2 + sizeof(int)*dimedge[0]*dimedge[1])*j,SEEK_SET);
			j=fread(workinglandscapes[k].ownerIDs,sizeof(int)*ncells,1,lf);
			if(!j) {Rprintf("Error reading landscape #%d, trying again.\n",j);k--;continue;}
			j=fread(workinglandscapes[k].tsd,sizeof(int)*ncells,1,lf);
			if(!j) {Rprintf("Error reading landscape #%d, trying again.\n",j);k--;continue;}
			j=fread(workinglandscapes[k].edge,sizeof(int)*dimedge[0]*dimedge[1],1,lf);
			if(!j) {Rprintf("Error reading landscape #%d, trying again.\n",j);k--;continue;}
#ifdef SHOWLANDSCAPES
			Rprintf("\n****************** NEW LANDSCAPE ************************\n");
			for(x=0,i=0;x<dim[1];x++) {
				for(y=0;y<dim[0];y++,i++) {					
					if(x<dim[1]/8 && y<dim[0]/1) Rprintf("%c",64+(int)(((float)(workinglandscapes[k].tsd[i])/(float)100)*48));
				}
				if(x<dim[1]/8) Rprintf("\n");
			}
			Rprintf("***************************************************\n");
#endif
		}
		fclose(lf);
}

static void permute(int *permutation,int n) {
    int i,j;
    for (i = 0; i < n; i++) {
        j=(int)((float)rand()/RAND_MAX*i);
        permutation[i] = permutation[j];
        permutation[j] = i;
    }
}

