#include "simulator.h"

void prepare(SEXP objfuncpars) {		// read global variables of this optimization run
	int i;
	SEXP tmp;
	progress=0;
	oriland=getListElement(objfuncpars,"landscape");
	pred=getListElement(objfuncpars,"models");
	rgama=getListElement(objfuncpars,"rgama");
//	landscapegamma=getListElement(objfuncpars,"landscapegamma");
	nyears=INTEGER_POINTER(getListElement(objfuncpars,"n.years"))[0];
	totalarea=INTEGER_POINTER(getListElement(objfuncpars,"total.area"))[0];
	intervals=getListElement(objfuncpars,"intervals");
	startyear=getListElement(objfuncpars,"startyear");
	trace=getListElement(objfuncpars,"trace");
	tmp=getListElement(objfuncpars,"savelandscapes");
	onelandscapeonly=getListElement(objfuncpars,"oneLandscapeOnly");
	savelandscapes=(tmp==R_NilValue ? 0 : (INTEGER_POINTER(tmp)[0] ? 1 : 0));
	minareaTSD=getListElement(objfuncpars,"minareaTSD");
	nlandscapes=INTEGER_POINTER(getListElement(objfuncpars,"n.landscape.repetitions"))[0];
	nrepeats=INTEGER_POINTER(getListElement(objfuncpars,"n.repeats"))[0];
	landscapefile=getListElement(objfuncpars,"landscape.file");				// filename where random landscapes are stored (optional)
	if(landscapefile!=R_NilValue) {
		FILE *lf=fopen(CHARACTER_VALUE(landscapefile),"r");
		if(!lf) error("Landscape file not found.");
		fclose(lf);
	}
	
	int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));					// nr. of cells in the landscape (count of ownerID field)
	int	*dimedge=INTEGER_POINTER(GET_DIM(VECTOR_ELT(oriland,2)));
	workinglandscapes=malloc(sizeof(WORKINGLANDSCAPE)*nlandscapes);
	for(i=0;i<nlandscapes;i++) {
		workinglandscapes[i].tsd=(int*)malloc(sizeof(int)*ncells);
		workinglandscapes[i].ownerIDs=(int*)malloc(sizeof(int)*ncells);
		workinglandscapes[i].edge=malloc(sizeof(int)*dimedge[0]*dimedge[1]);
	}

	// compute random uniform distributions of TSD values to initialize landscapes
	short max_ano=INTEGER_POINTER(GET_DIM(pred))[0]-1;		// maximum age for which there are model predictions (e.g. biomass values)
	for(i=0;i<MAXNOWNERS;i++) unifTSD[i]=(int)((float)rand()/RAND_MAX*max_ano);
}

void cleanUp() {
	int i;
	for(i=0;i<nlandscapes;i++) {
		free(workinglandscapes[i].tsd);
		free(workinglandscapes[i].ownerIDs);
		free(workinglandscapes[i].edge);
	}

	free(workinglandscapes);
}

SEXP loadLandscape(SEXP n) {
	int i,k,j,totallands,*tmp;
	int ncells=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));
	int	*dimedge=INTEGER_POINTER(GET_DIM(VECTOR_ELT(oriland,2)));
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(VECTOR_ELT(oriland,1),1)));	// landscape dimensions
	FILE *lf=fopen(CHARACTER_VALUE(landscapefile),"r");
	WORKINGLANDSCAPE lnd;
	SEXP ret,oid,oc;
	int *tmpt1,*tmpt2,*tmpt3;
	
	lnd.tsd=(int*)malloc(sizeof(int)*ncells);
	lnd.ownerIDs=(int*)malloc(sizeof(int)*ncells);
	lnd.edge=malloc(sizeof(int)*dimedge[0]*dimedge[1]);
	
	j=fread(&totallands,sizeof(int),1,lf);
	if(!j) error("Error #1 reading landscape file.");
	Rprintf("%d landscapes found in file.\n",totallands);

	j=INTEGER_POINTER(n)[0];
	Rprintf("Reading landscape #%d... ",j);
	
	tmpt1=lnd.ownerIDs;		// pointers were stored in file for comodity, but are meaningless. save malloc'ed pointers before overriding values
	tmpt2=lnd.tsd;
	tmpt3=lnd.edge;
			
	fseek(lf,sizeof(int)+sizeof(WORKINGLANDSCAPE)*j,SEEK_SET);
	i=fread(&lnd,sizeof(WORKINGLANDSCAPE),1,lf);
	Rprintf("%d owners.\n",lnd.nowners);

	lnd.ownerIDs=tmpt1;
	lnd.tsd=tmpt2;
	lnd.edge=tmpt3;

	fseek(lf,sizeof(int)+sizeof(WORKINGLANDSCAPE)*totallands+(sizeof(int)*ncells*2 + sizeof(int)*dimedge[0]*dimedge[1])*j,SEEK_SET);
	j=fread(lnd.ownerIDs,sizeof(int)*ncells,1,lf);
	if(!j) error("Error #2 reading landscape file.");
	j=fread(lnd.tsd,sizeof(int)*ncells,1,lf);
	if(!j) error("Error #3 reading landscape file.");
	j=fread(lnd.edge,sizeof(int)*dimedge[0]*dimedge[1],1,lf);
	if(!j) error("Error #4 reading landscape file.");
#ifdef SHOWLANDSCAPES
			Rprintf("\n****************** NEW LANDSCAPE ************************\n");
			for(x=0,i=0;x<dim[1];x++) {
				for(y=0;y<dim[0];y++,i++) {					
					if(x<dim[1]/8 && y<dim[0]/1) Rprintf("%c",64+(int)(((float)(lnd.tsd[i])/(float)100)*48));
				}
				if(x<dim[1]/8) Rprintf("\n");
			}
			Rprintf("***************************************************\n");
#endif
	fclose(lf);
		
	
	PROTECT(ret=NEW_LIST(2));
	PROTECT(oid=allocMatrix(INTSXP,dim[0],dim[1]));
	PROTECT(oc=allocMatrix(INTSXP,MAXOWNEDCELLS,lnd.nowners));
	SET_VECTOR_ELT(ret,0,oid);
	SET_VECTOR_ELT(ret,1,oc);
	memcpy(INTEGER_POINTER(oid),lnd.ownerIDs,sizeof(int)*dim[0]*dim[1]);
	for(i=0;i<lnd.nowners;i++) {
		tmp=INTEGER_POINTER(oc)+i*MAXOWNEDCELLS;
		for(j=0;j<lnd.ownercells[i].ncells;j++) {		
			*tmp=(int)lnd.ownercells[i].ownedcells[j];
			tmp++;
		}
	}

	free(lnd.tsd);
	free(lnd.ownerIDs);
	free(lnd.edge);
	UNPROTECT(3);
	return(ret);
}
