#include "simulator.h"

typedef struct {
	int r,c,nneigh;
} CELL;

static int checkNeighbors(int r,int c,int patchcounter,CELL *patch,int m_nrows,int *tid);

// hg: hexgrid
// rgama: vector with N values of a gamma distribution
void randLandscape(SEXP hg,SEXP rgama,int *pout) {
	SEXP	edge=VECTOR_ELT(hg,2);
	int		*id=INTEGER_POINTER(VECTOR_ELT(hg,0));
	int		*dimed=INTEGER_POINTER(GET_DIM(edge));
	int		*dimid=INTEGER_POINTER(GET_DIM(VECTOR_ELT(hg,0)));
	int		*pgama=INTEGER_POINTER(rgama);
	short	lgama=LENGTH(rgama);
	int		nelem=dimed[0]*dimed[1],buf=nelem*sizeof(int),bufshort=nelem*sizeof(short),ncells=0;
	register int	i,j,k,l,x,r,c,z;
	int		counter,gamacounter=0,area,areacounter;
	CELL	possible[MAXPOSSIBLE],*pposs[MAXPOSSIBLE];			
	int		possiblecounter;
	int		lenged=dimed[0]*dimed[1],dimed2=2*dimed[0];
	short	*pindex,pposscounter;
	
	int		m_nrows,m_ncols;
	int		*tid,patchcounter;
	CELL	patch[MAXPATCH];

	tid=(int*)malloc(buf);
	pindex=(short*)malloc(bufshort);
	if(!tid) error("out of memory.");
	if(!pindex) error("out of memory.");
	
	m_nrows=dimed[0];m_ncols=dimed[1];
	
	memcpy(tid,INTEGER_POINTER(edge),buf);		// copy edges
	memset(pout,0,buf);

	for(i=0,k=0;i<m_nrows;i+=2,k++) {					// copy IDs to within edges
		for(j=0,l=k;j<lenged;j+=dimed2,l+=dimid[0]) {
			z=j+i;
			tid[z]=id[l];
			pout[z]=tid[z];
			if(tid[z]!=NA_INTEGER) ncells++;				// count free cells
		}
	}

	counter=1;									// landowner counter
	do {										// for each owner one patch		
		patchcounter=0;
		patch[0].nneigh=0;
		z=(int)((float)rand()/((float)RAND_MAX+1)*ncells);
		k=0;x=0;
		for(j=0;j<m_ncols;j+=2,k+=m_nrows-1) {		// find a random free cell
			l=k+m_nrows;
			for(;k<l;k+=2) {
				if(tid[k]!=NA_INTEGER) {
					if(x==z) goto a1;			// found free cell
					x++;
				}
			}
		}		
a1:	
		r=k % m_nrows;
		c=(int)k/m_nrows;
		area=pgama[gamacounter];
		gamacounter++;
		if(gamacounter>=lgama) gamacounter=0;
		areacounter=0;
		
		do {
			z=c*m_nrows+r;
			pout[z]=-counter;						// assign owner to cell r,c
			tid[z]=NA_INTEGER;						// cell is not free any more
			ncells--;								// decrease free cell counter

			patch[patchcounter].r=r;					// add to owner index this cell
			patch[patchcounter].c=c;
			patchcounter++;
#ifdef ERRORCHECK
			if(patchcounter>=MAXPATCH) error("you have to recompile and increase MAXPATCH");
#endif
			possiblecounter=0;
			memset(pindex,0,bufshort);
			for(i=0;i<patchcounter;i++) {			// run through all cells of one owner
				if(patch[i].nneigh==6) continue;	// no free contiguous cells, skip.
				r=patch[i].r;
				c=patch[i].c;
				
				if(c<m_ncols-2) {
					if(tid[(c+1)*m_nrows+r]==-1 && tid[(c+2)*m_nrows+r]!=NA_INTEGER) {
						possible[possiblecounter].r=r;
						possible[possiblecounter].c=c+2;
						pindex[(c+2)*m_nrows+r]++;
						possiblecounter++;
#ifdef ERRORCHECK						
						if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
					}
					if(r>1) {
						if(tid[(c+1)*m_nrows+r-1]==-3 && tid[(c+2)*m_nrows+r-2]!=NA_INTEGER) {
							possible[possiblecounter].r=r-2;
							possible[possiblecounter].c=c+2;
							pindex[(c+2)*m_nrows+r-2]++;
							possiblecounter++;
#ifdef ERRORCHECK						
							if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
						}
					}
					if(r<m_nrows-2) {
						if(tid[(c+1)*m_nrows+r+1]==-4 && tid[(c+2)*m_nrows+r+2]!=NA_INTEGER) {
							possible[possiblecounter].r=r+2;
							possible[possiblecounter].c=c+2;
							pindex[(c+2)*m_nrows+r+2]++;
							possiblecounter++;										
#ifdef ERRORCHECK						
							if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
						}
					}
				}

				if(c>1) {
					if(tid[(c-1)*m_nrows+r]==-1 && tid[(c-2)*m_nrows+r]!=NA_INTEGER) {
						possible[possiblecounter].r=r;
						possible[possiblecounter].c=c-2;
						pindex[(c-2)*m_nrows+r]++;
						possiblecounter++;										
#ifdef ERRORCHECK						
						if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
					}
					if(r>1) {
						if(tid[(c-1)*m_nrows+r-1]==-4 && tid[(c-2)*m_nrows+r-2]!=NA_INTEGER) {
							possible[possiblecounter].r=r-2;
							possible[possiblecounter].c=c-2;
							pindex[(c-2)*m_nrows+r-2]++;
							possiblecounter++;										
#ifdef ERRORCHECK						
							if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
						}
					}
					if(r<m_nrows-2) {
						if(tid[(c-1)*m_nrows+r+1]==-3 && tid[(c-2)*m_nrows+r+2]!=NA_INTEGER) {
							possible[possiblecounter].r=r+2;
							possible[possiblecounter].c=c-2;
							pindex[(c-2)*m_nrows+r+2]++;
							possiblecounter++;										
#ifdef ERRORCHECK						
							if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
						}
					}
				}
				if(r>1) {
					if(tid[c*m_nrows+r-1]==-2 && tid[c*m_nrows+r-2]!=NA_INTEGER) {
						possible[possiblecounter].r=r-2;
						possible[possiblecounter].c=c;
						pindex[c*m_nrows+r-2]++;
						possiblecounter++;										
#ifdef ERRORCHECK						
						if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
					}
				}
				if(r<m_nrows-2) {
					if(tid[c*m_nrows+r+1]==-2 && tid[c*m_nrows+r+2]!=NA_INTEGER) {
						possible[possiblecounter].r=r+2;
						possible[possiblecounter].c=c;
						pindex[c*m_nrows+r+2]++;
						possiblecounter++;										
#ifdef ERRORCHECK						
						if(possiblecounter>=MAXPOSSIBLE) error("you have to recompile and increase MAXPOSSIBLE");
#endif
					}
				}
			}
			
			if(possiblecounter==0) break;
			if(possiblecounter==1) {
				r=possible[0].r;
				c=possible[0].c;				
				patch[patchcounter].nneigh=checkNeighbors(r,c,patchcounter,patch,m_nrows,tid);		// adjust neighbor counts
			} else {
#if FRACTALDIM>0				
				pposscounter=0;
				for(i=0;i<possiblecounter;i++) {
					if(pindex[possible[i].c*m_nrows+possible[i].r]>FRACTALDIM) {	// cell was requested more than FRACTALDIM times, include it as candidate
						pposs[pposscounter]=&possible[i];
						pposscounter++;
					}
				}
				if(!pposscounter) {		// no candidates
#endif					
					z=(int)((float)rand()/((float)RAND_MAX+1)*possiblecounter);
					r=possible[z].r;
					c=possible[z].c;
#if FRACTALDIM>0					
				} else {
					z=(int)((float)rand()/((float)RAND_MAX+1)*pposscounter);
					r=(*pposs[z]).r;
					c=(*pposs[z]).c;					
				}
#endif				
				patch[patchcounter].nneigh=checkNeighbors(r,c,patchcounter,patch,m_nrows,tid);		// adjust neighbor counts
			}
			areacounter++;
			if(areacounter>=area) break;
		} while(TRUE);	
		counter++;
	} while(ncells>0);

	free(tid);
	free(pindex);
}

static int checkNeighbors(int r,int c,int patchcounter,CELL *patch,int m_nrows,int *tid) {
	int i,j=0;
	for(i=0;i<patchcounter-1;i++) {
		if(r==patch[i].r) {
			if(c==patch[i].c-2) {
				if(tid[(c+1)*m_nrows+r]==-1) {patch[i].nneigh++;j++;continue;}
			}
			if(c==patch[i].c+2) {
				if(tid[(c-1)*m_nrows+r]==-1) {patch[i].nneigh++;j++;continue;}
			}
		}
		if(c==patch[i].c) {
			if(r==patch[i].r-2) {
				if(tid[c*m_nrows+r+1]==-2) {patch[i].nneigh++;j++;continue;}
			}

			if(r==patch[i].r+2) {

				if(tid[c*m_nrows+r-1]==-2) {patch[i].nneigh++;j++;continue;}
			}
		}
		
		if(c==patch[i].c-2) {
			if(r==patch[i].r-2) {
				if(tid[(c+1)*m_nrows+r+1]==-4) {patch[i].nneigh++;j++;continue;}
			}
			if(r==patch[i].r+2) {
				if(tid[(c+1)*m_nrows+r-1]==-3) {patch[i].nneigh++;j++;continue;}
			}
		}
		if(c==patch[i].c+2) {
			if(r==patch[i].r-2) {
				if(tid[(c-1)*m_nrows+r+1]==-3) {patch[i].nneigh++;j++;continue;}
			}
			if(r==patch[i].r+2) {
				if(tid[(c-1)*m_nrows+r-1]==-4) {patch[i].nneigh++;j++;continue;}
			}
		}
	}
	return(j);
}

