#define	ERRORCHECK					// do some basic checks for overflow errors
//#define EVENYEARS					// only compute objectives every other year
//#define KEEPSTRATEGYGROUPS			// force algorithm to maintain the given nr. of groups of the strategy, or allow solutions of N groups to degrade to N-1 groups?
#define NONAS						// assume there are no NAs in the landscape (i.e. rectangular landscape)
#define ONEMODELONLY				// only one biomass accumulation curve
//#define VERBOSE
//#define SHOWLANDSCAPES

#define MAXNOWNERS		2700		// maximum number of owners in the landscape, increase and recompile if needed
#define MAXOWNEDCELLS	1000
// NOTE: if you change MAXNOWNERS or MAXOWNEDCELLS, you must rebuild the landscape files
#define MAXCORETHREADS	20
#define MAXTRACEYEARS	150
#define MAXPARS			12

// for landscape generator:
#define		MAXPATCH	6000
#define		MAXPOSSIBLE	3000
#define		FRACTALDIM	1

#ifndef SIMULATOR_H
	#define SIMULATOR_H
	#include <R.h>
	#include <Rdefines.h>
	#include <pthread.h>
	#include <R_ext/Utils.h>
	
	void randLandscape(SEXP hg,SEXP rgama,int *pout);
	SEXP getListElement(SEXP list, const char *str);
	inline double max(double a,double b) {if(a>b) return a; else return b;}
	void *simulator(void *thrdata);
	void loadLandscapesFromFile();
	void objectiveFunction(double **par,double **out,SEXP objfuncpars, SEXP);
	void cleanUp();
	void prepare(SEXP objfuncpars);
	SEXP loadLandscape(SEXP n);
	void constraintFunction(double **par,double **out,SEXP objfuncpars);
	
	typedef struct {
		unsigned short ownedcells[MAXOWNEDCELLS];	// array where owned cells (their indexes) are stored
		int ncells;			// number of owned cells
	} OWNERCELLS;

	typedef struct {		// structure to hold data for one owner
		short group;		// owner group
		short interval;		// desired management interval
		short nextmanyear;	// next year when management will actually be done
	} OWNER;

	typedef struct {
		int *ownerIDs,*tsd,*edge;
		int nowners;
		OWNERCELLS ownercells[MAXNOWNERS];
		int ownerindexes[MAXNOWNERS];
	} WORKINGLANDSCAPE;

	typedef struct {
		int *manint,ready,returned;
		OWNER ownerlist[MAXNOWNERS];
		WORKINGLANDSCAPE *workinglandscape;
		SEXP tracelist;
		int *tsd;
		double *pobj;
	} THREADDATA;
	WORKINGLANDSCAPE *workinglandscapes;
	
	THREADDATA td[MAXCORETHREADS];
	int nyears,totalarea,unifTSD[MAXNOWNERS],nlandscapes,nrepeats,progress;
	SEXP oriland,pred,rgama,intervals,startyear,trace,minareaTSD,landscapegamma,landscapefile,onelandscapeonly;
	int savelandscapes;
#endif


