# some general functions for plotting and analysing results
p=palette()
p[5]="magenta"
p[6]="orange"
palette(p)

visualize=function(r,norm="euclidean",highlight,disp="objectives",ranges,limits=NULL,dispminimum=switch(disp,o=1,p=1:10),legend=T,color=T,title=NULL,letter,objectives=1:dim(r$value)[2]) {
    # compute distance to the ideal point (Blasco et al.)
    normv=compute.distance(r,norm)

    # transform just for plotting purposes. converts raw objective values (used in optimization) into human-readable values
	r=transformObjectives(r)
	
	# reduce solution set to those that fulfill restrictions
    subsetted.results=r
    subsetted.normv=normv
	if(!is.null(limits)) {
		ssr=subset(r,limits)
		highlight=ssr$which.solutions
		subsetted.results=ssr$subsetted.results
		subsetted.normv=subsetted.normv[highlight]
    }

    ylim=range(normv)
    retpars=list()
    pars=process.parameters(r)
	subsetted.pars=process.parameters(subsetted.results)

	cat("Number of solutions in each strategy:")
	print(table(get.n.groups(pars)))

    switch(pmatch(disp,c("objectives","parameters")),{	# plot optimization objectives
        dir.objectives=c(-1,-1,1,1,1,1)
	    npars=get.n.groups(pars)
	    n.groups=sort(unique(get.n.groups(subsetted.pars)))
	    if(length(objectives)>1) {
	    	nobj=dim(r$value)[2]
			layout(best.layout(nobj,nc=2))
		}
		par(mai=c(0.3,0.3,0.2,0.1),mgp=c(1.1,0.3,0),tcl=-0.3,cex.lab=1.3,cex.axis=1.1)
		
		for(i in objectives) {		# for each objective (each plot)
			plot.new()
			if(missing(ranges)) {
				plot.window(xlim=range(r$value[,i]),ylim=ylim)
			} else {
				if(is.null(ranges[[i]]))
					plot.window(xlim=range(r$value[,i]),ylim=ylim)
				else
					plot.window(xlim=ranges[[i]],ylim=ylim)
			}
			leg=character()
			legcol=numeric()

			if(!is.null(limits)) {	# draw vertical lines in the limits
				if(length(limits[[i]])==2) {
					usr=par("usr")
					rect(usr[1],usr[3],limits[[i]][1],usr[4],col="#eeeeee",border=NA)
					rect(limits[[i]][2],usr[3],usr[2],usr[4],col="#eeeeee",border=NA)
					abline(v=limits[[i]],lty=2,lwd=1);
				}
			}
			
			# plot all points
			if(color) c=npars else c="gray"
			points(r$value[,i],normv,col=c,pch=ifelse(c==1,21,19),cex=ifelse(c==1,1.7,0.6),lwd=ifelse(c==1,2,1))

			# and highlight minimum values
			if(!is.na(dispminimum[1])) {
				if(any(dispminimum!=F)) {
					m=order(normv)[dispminimum]
					points(x=r$value[m,i],y=normv[m],col=npars[m],cex=2,pch=17)
				}
			}

			# highlight specific points
			if(!missing(highlight)) {
				m=highlight
				if(length(m)>0) {
					points(r$value[m,i],normv[m],col="black",cex=1.4,pch=21,bg=npars[m],lwd=0.5)
					if(i==1) {
						for(tmpj in m) retpars=c(retpars,list(pars[tmpj,]))
					}
				}
			}		
			
			adjustarrow=0
			if(legend && length(n.groups)>0) {
				if(i==1) {	# only draw legend in the first plot
					if(color)
						legend("bottomright",leg=paste(n.groups,"-group str.",sep=""),col=n.groups,title=NA,bty="n",pch=ifelse(n.groups==1,21,19),pt.cex=ifelse(n.groups==1,1.5,1))
					else
						legend("bottomright",leg=paste(n.groups,"-group str.",sep=""),col="black",title=NA,bty="n",pch=21,pt.cex=1.4,pt.bg=n.groups)
					adjustarrow=0.3
				}
			}

			# draw arrows telling the direction that is preferrable (max or min)
			usr=par("usr")
			if(dir.objectives[i]>0)
				arrows(usr[1]+(usr[2]-usr[1])*(0.9-adjustarrow),usr[3]+(usr[4]-usr[3])*0.05,usr[1]+(usr[2]-usr[1])*(0.99-adjustarrow),length=0.1,lwd=1.4)
			else
				arrows(usr[1]+(usr[2]-usr[1])*(0.99-adjustarrow),usr[3]+(usr[4]-usr[3])*0.05,usr[1]+(usr[2]-usr[1])*(0.9-adjustarrow),length=0.1,lwd=1.4)

			axis(1)
			axis(2,las=2,mgp=c(1.1,0.6,0))
			box(bty="L",lwd=2)
			abline(h=pretty(ylim),lty=3,col="grey",lwd=0.75)
			title(xlab=colnames(r$value)[i])
			if(!missing("title")) title(main=title)
			if(missing("letter"))
				mtext(LETTERS[i],3,adj=0,cex=1,font=2)
			else
				mtext(letter,3,adj=0,cex=1,font=2)
		}
		
		if(length(retpars)>0) {
			cat("Highlighted solutions:\n",sapply(retpars,FUN=function(a) {
				paste(format.pars(a),"\n",sep="")
			}))
			return(invisible(retpars))
		} else {	# find best solutions in all scenarios
			m=order(normv)[1:10]
			for(tmpj in m) retpars=c(retpars,list(pars[tmpj,]))
			cat("Overall best solutions:\n",sapply(retpars,FUN=function(a) {
				paste(format.pars(a),"\n",sep="")
			}))
			return(invisible(retpars))
		}
    },{										# plot optimization parameters
    	npars=get.n.groups(subsetted.pars)
    	n.groups=sort(unique(npars))
    	values=matrix(nc=4,nr=0)
    	colnames(values)=c("Management interval (yr)","Proportion of owners","Scenario","Dim")
    	
    	if(length(subsetted.normv)==0)
    		best=NULL
    	else {
			if(is.na(dispminimum[1]))
				best=1:length(subsetted.normv)
			else
				best=order(subsetted.normv)[dispminimum]
		}

		par(mar=c(2.5,2.5,1.5,0.5),mgp=c(1.3,0.4,0),cex.lab=1.3,cex.main=1.3,cex.axis=1.1)
		plot.new()
		plot.window(xlim=c(0,100),ylim=c(0,1))
    	
    	if(length(best)>0) {
			cat("\nPlotted solutions:\n")
	    	cat(apply(subsetted.pars[best,,drop=F],1,FUN=function(a) {
				paste(format.pars(a),"\n",sep="")
			}))
			intervals=subsetted.pars[best,1:5,drop=F]
			proportions=subsetted.pars[best,6:10,drop=F]
			for(k in 1:dim(intervals)[1]) {		# increasing order
				ord=order(intervals[k,])
				intervals[k,]=floor(intervals[k,][ord])
				proportions[k,]=proportions[k,][ord]
			}
		
			for(k in 1:dim(intervals)[1]) {
				values=rbind(values,cbind(c(intervals[k,],NA),c(proportions[k,],NA),sum(!is.na(intervals[k,])),0))#dim.solutions[k]))
			}

			d=dim(values)[1]
			if(length(best)<30) {
				segments(values[1:(d-1),1],values[1:(d-1),2],values[2:d,1],values[2:d,2],col=ifelse(values[1:(d-1),"Dim"],NA,values[,"Scenario"]),lty=3)
				points(values[,1],values[,2],col=paste(rgb(t(col2rgb(values[,"Scenario"])),max=255),ifelse(values[,"Dim"],"44","ff"),sep=""),pch=21,lwd=2,cex=ifelse(values[,"Dim"],0.7,1))
	   		} else points(values[,1],values[,2],col=paste(rgb(t(col2rgb(values[,"Scenario"])),max=255),ifelse(values[,"Dim"],"44","ff"),sep=""),pch=21,cex=ifelse(values[,"Dim"],0.7,0.7))

			if(length(best)<30) {
				rug(values[,2],side=2,ticksize=0.015)
				rug(values[,1],side=1,ticksize=0.015)
			}
		} else {
			legend=F
			usr=par("usr")
			text((usr[2]-usr[1])/2,(usr[4]-usr[3])/2,"(no solutions)",adj=c(0.5,0.5))
		}
		axis(1,at=c(seq(0,80,20),99),labels=c(seq(0,80,20),"No-Man."))
		axis(2)
		box(bty="l",lwd=2)
		axis(1,at=seq(0,90,10),labels=F,tck=1,lty=3,col.ticks="gray")
		axis(1,at=99,labels=F,tck=1,lty=3,col.ticks="black",lwd=2)
		axis(2,at=seq(0.1,1,0.1),labels=F,tcl=-0.2)
    	title(main=title,xlab=colnames(values)[1],ylab=colnames(values)[2])
    	if(!missing("letter")) mtext(letter,3,adj=0,cex=1,font=2)
    	if(legend) legend("top",leg=paste(n.groups,"-group str.",sep=""),col=n.groups,bty="n",pch=21,pt.lwd=2);
    })
}

transformObjectives=function(r) {	# transform objectives just for plotting purposes. converts raw objective values (used in optimization) into human-readable values
    r$value[,1]=r$value[,1]*100
    r$value[,3]=r$value[,3]*-1
    r$value[,4]=r$value[,4]*-1
    for(j in 5:dim(r$value)[2]) {
    	r$value[,j]=-r$value[,j]*otherpars$minareaTSD[j-4]*100
    }
    return(r)
}

plotObjectiveMatrix=function(r,limits=NULL,ranges) {
    colors.objectives=c(5,2,2,2,4,4)
    dir.objectives=c(-1,-1,1,1,1,1)
    nobj=dim(r$value)[2]
    pares=combn(1:nobj,2)
    cnam=colnames(r$value)
    
    r=transformObjectives(r)
    
	subsetted.results=r
    if(!is.null(limits)) {
    	cex=1
		ssr=subset(r,limits)
		highlight=ssr$which.solutions
		subsetted.results=ssr$subsetted.results
    } else cex=0.2
    pars=process.parameters(subsetted.results)
    npars=get.n.groups(pars)
    n.groups=sort(unique(npars))
    print(apply(pars,1,format.pars))
    
    par(mai=c(0.3,0.3,0.2,0.1),mgp=c(1.1,0.3,0),lwd=2,tcl=-0.3,cex.lab=1.1)
    layout(best.layout(dim(pares)[2]))
    for(i in 1:dim(pares)[2]) {           
    	plot.new()
		if(missing(ranges)) {
			xlim=range(r$value[,pares[1,i]])
			ylim=range(r$value[,pares[2,i]])
		} else {
			if(is.null(ranges[[pares[1,i]]]))
				xlim=range(r$value[,pares[1,i]])
			else
				xlim=ranges[[pares[1,i]]]
			
			if(is.null(ranges[[pares[2,i]]]))
				ylim=range(r$value[,pares[2,i]])
			else
				ylim=ranges[[pares[2,i]]]
		}
		plot.window(xlim=xlim,ylim=ylim)

		usr=par("usr")
		if(!is.null(limits)) {	# draw panes representing the restrictions
			if(length(limits[[pares[1,i]]])==2) {
				rect(usr[1],usr[3],limits[[pares[1,i]]][1],usr[4],col="#eeeeee",border=NA)
				rect(limits[[pares[1,i]]][2],usr[3],usr[2],usr[4],col="#eeeeee",border=NA)
			}
			
			if(length(limits[[pares[2,i]]])==2) {
				rect(usr[1],usr[3],usr[2],limits[[pares[2,i]]][1],col="#eeeeee",border=NA)
				rect(usr[1],limits[[pares[2,i]]][2],usr[2],usr[4],col="#eeeeee",border=NA)
				abline(h=limits[[pares[2,i]]],lty=2,lwd=1)
			}
		}
		
		# draw all points in gray
    	points(r$value[,pares[1,i]],r$value[,pares[2,i]],pch=19,col="gray",cex=0.2)

		# draw lines in the panes representing the restrictions
		if(!is.null(limits)) {
			if(length(limits[[pares[2,i]]])==2) abline(h=limits[[pares[2,i]]],lty=2,lwd=1)
			if(length(limits[[pares[1,i]]])==2) abline(v=limits[[pares[1,i]]],lty=2,lwd=1)
		}
   	
   		# draw highlighted points in colors
    	points(subsetted.results$value[,pares[1,i]],subsetted.results$value[,pares[2,i]],col=npars,cex=ifelse(npars==1,1.5,cex+0.1),pch=ifelse(npars==1,21,19),lwd=ifelse(npars==1,2,1),bg="white")

		# draw an asterisk telling the direction that is preferrable for both axes (max or min)
		adj=c(0,0)
		if(dir.objectives[pares[1,i]]>0) {x=usr[2];adj[1]=1} else {x=usr[1];adj[1]=0}
		if(dir.objectives[pares[2,i]]>0) {y=usr[4];adj[2]=1} else {y=usr[3];adj[2]=0.5}
		text(x,y,labels="*",cex=4,adj=adj)
				
    	axis(1)
    	axis(2)
    	box(bty="l")
    	title(xlab=cnam[pares[1,i]],col.lab=colors.objectives[pares[1,i]])
    	title(ylab=cnam[pares[2,i]],col.lab=colors.objectives[pares[2,i]])
    	mtext(LETTERS[i],3,adj=0,cex=1,font=2)

		if(i==1) legend("topright",leg=paste(n.groups,"-group str.",sep=""),col=n.groups,bty="n",pch=ifelse(n.groups==1,21,19),pt.cex=ifelse(n.groups==1,1.3,0.8));
    }
}

ordinateSolutions=function(par,title,letter,nominalset,xlim=c(-0.5,0.5),ylim=c(-0.5,0.5)) {
	par(mar=c(2.5,2.5,1.5,0.8),mgp=c(1.3,0.4,0),lwd=2,tcl=-0.3,cex.main=1.3,cex.lab=1.3,cex.axis=1.1)
	
	# plot PCoA of solutions using distance matrix computed from the % intersection of density functions of management intervals
	if(any(class(par)!="matrix")) {
		par1=matrix(NA,nr=0,nc=dim(par[[1]]$par)[2])
		for(i in 1:length(par)) {
			par1=rbind(par1,par[[i]]$par)
		}
		color=rep(1:length(par),times=sapply(par,function(a) dim(a$par)[1]))
		par=par1
		cat("Total number of solutions:",length(color),"\n")
		flush.console()
	} else color=get.n.groups(process.parameters(par))
	
	dens=get.density(process.parameters(par))
	nsols=dim(dens)[2]
	densfunc=as.list(vector(len=nsols))
	for(sol in 1:nsols) {
		densfunc[[sol]]=density(dens[,sol,1],wei=dens[,sol,2],bw=5,from=-49,to=150,n=200)
	}
	dist=matrix(NA,nc=nsols,nr=nsols)
	for(i in 1:nsols) {
		for(j in i:nsols) {
			d=1-sum(pmin(densfunc[[i]]$y,densfunc[[j]]$y))
			dist[i,j]=d
			dist[j,i]=d
		}
	}
	dist[dist<0]=0
	pcoa=cmdscale(dist,k=2,eig=T)
	eig=round(pcoa$eig/sum(pcoa$eig)*100,1)
	
	plot(pcoa$points,pch=21,col="black",cex=1.15,bg=color,lwd=0.3,xlab=paste("PCoA1 (",eig[1],"%)",sep=""),ylab=paste("PCoA2 (",eig[2],"%)",sep=""),bty="l",xlim=xlim,ylim=ylim)
	if(!missing("nominalset")) points(pcoa$points[color==nominalset,,drop=F],pch=24,col="black",bg=nominalset,lwd=1.5,cex=1.5)
		
	abline(h=0,v=0,lty=3,col="gray",lwd=1)
	if(!missing("letter")) mtext(letter,3,adj=0,cex=1,font=2)
	if(!missing("title")) title(main=title)	
	return(dist)
}


plotAreas=function(TSD,range.years.simul=c(1,dim(TSD[[1]])[1]),inter=c(0,10,20,30,60,100),file="areas" ) {
	bl=best.layout(length(TSD))
	openDevice(output,name=file,wid=3.5*dim(bl)[2],hei=3.5*dim(bl)[1])
	layout(bl)
	par(mar=c(2,2,1,0.5),mgp=c(1.1,0.3,0),lwd=2,tcl=-0.3,cex.lab=1.1,cex.axis=1)
	for(i in 1:length(TSD)) {
	    meantsd=apply(TSD[[i]],c(1,2),mean);
	    sdtsd=apply(TSD[[i]],c(1,2),sd);

	    plot.new();plot.window(xlim=c(1,dim(TSD[[i]])[1]),ylim=c(0,1));
	    for(j in 1:dim(TSD[[i]])[2]) {
			barrasx=seq(range.years.simul[1],range.years.simul[2]-10,by=10)+j*2;
			#lines(range.years.simul[1]:range.years.simul[2],meantsd[range.years.simul[1]:range.years.simul[2],j],col="gray",lwd=1.5)
		
			lines(range.years.simul[1]:range.years.simul[2],meantsd[range.years.simul[1]:range.years.simul[2],j],col=j,lwd=1.3)
			segments(barrasx,meantsd[barrasx,j]+sdtsd[barrasx,j],barrasx,meantsd[barrasx,j]-sdtsd[barrasx,j],lwd=0.8,col="black");
			segments(barrasx-1,meantsd[barrasx,j]+sdtsd[barrasx,j],barrasx+1,meantsd[barrasx,j]+sdtsd[barrasx,j],lwd=0.8,col="black");
			segments(barrasx-1,meantsd[barrasx,j]-sdtsd[barrasx,j],barrasx+1,meantsd[barrasx,j]-sdtsd[barrasx,j],lwd=0.8,col="black");
	    }
	    axis(1)
	    axis(2,at=seq(0,100,by=10)/100,lab=paste(seq(0,100,by=10),sep=""))
	    abline(h=0.1,lty=3,col="gray")
	    
	    main=sapply(strsplit(names(TSD[i])," ",fixed=T),function(a) {format.pars(as.numeric(a)[-length(a)]) })
	    #title(main=format.pars(pars[[i]]),ylab="% of landscape area",xlab="Year");
	    title(main=main,ylab="% of landscape area",xlab="Year",cex.main=1)
	    box(bty="l")
	    legend("topleft",leg=paste(inter[1:(length(inter)-1)],"-",inter[2:length(inter)]," yr",sep=""),lty=1,bty="n",cex=0.8,col=1:length(inter),lwd=1.3)
	}
	dev.off()
}
plotObjectives=function(objectives,range.years.simul=c(1,dim(objectives[[1]])[1] ),obj.names=NULL,legend) {
	meanobj=lapply(objectives,function(x) {apply(x,c(1,2),mean,na.rm=T)})
	sdobj=lapply(objectives,function(x) {apply(x,c(1,2),sd,na.rm=T)})
	ran=lapply(meanobj,function(x) {
		apply(x[range.years.simul[1]:range.years.simul[2],],2,range,na.rm=T)
	});
	
	rang=matrix(0,nc=2,nr=0);
	for(i in 1:dim(ran[[1]])[2]) rang=rbind(rang,range(unlist(lapply(ran,"[",,i))));
	bl=best.layout(dim(meanobj[[1]])[2],nc=2)
	openDevice(output,name="objectives",wid=3.5*dim(bl)[2],hei=3.5*dim(bl)[1])
	layout(bl)
	par(mai=c(0.3,0.2,0.2,0.1),mgp=c(1.1,0.3,0),tcl=-0.3,cex=0.7,cex.lab=1.3,cex.axis=1.1)

	#N=(sapply(pars,length)+1)/2
	for(i in 1:dim(ran[[1]])[2]) {
	    plot.new();
	    if(i>=5) rang[i,1]=0;	# for these, minimum is always zero
	    plot.window(xlim=c(1,dim(meanobj[[1]])[1]),ylim=rang[i,]);
	    
	    for(j in 1:length(meanobj)) {
			barrasx=seq(range.years.simul[1],range.years.simul[2]-10,by=10)+j*2;
			lines(range.years.simul[1]:range.years.simul[2],meanobj[[j]][range.years.simul[1]:range.years.simul[2],i],lwd=1.5,col=j)
			segments(barrasx,meanobj[[j]][barrasx,i]-sdobj[[j]][barrasx,i],barrasx,meanobj[[j]][barrasx,i]+sdobj[[j]][barrasx,i],col=j,lwd=1.4)
			segments(barrasx-1,meanobj[[j]][barrasx,i]-sdobj[[j]][barrasx,i],barrasx+1,meanobj[[j]][barrasx,i]-sdobj[[j]][barrasx,i],col=j,lwd=1.4)
			segments(barrasx-1,meanobj[[j]][barrasx,i]+sdobj[[j]][barrasx,i],barrasx+1,meanobj[[j]][barrasx,i]+sdobj[[j]][barrasx,i],col=j,lwd=1.4)
	    }
	    
	    if(i==1) {
	    	if(!missing(legend)) legend()
	    }
	    axis(1);axis(2);
	    title(main=obj.names[i],xlab="Year",cex.main=1.3)
	    if(i>=5) abline(h=10,lty=2)
	    if(i!=1) abline(v=20,lty=2,col="gray")
	    box(bty="l",lwd=2)
	    mtext(LETTERS[i],3,adj=0,cex=1,font=2)
	}
	dev.off()
}

plotUncertainty=function(rgama,which.boxes=seq(5,95,by=5)) {
	par(mai=c(0.45,0.45,0.1,0.1),mgp=c(1.2,0.35,0),tcl=-0.3)
	boxplot(rgama[,which.boxes],range=0,at=which.boxes,names=which.boxes,boxwex=3,xlab="Proposed management interval",ylab="Realized management interval",outcex=0.6,lwd=0.7)
	abline(0,1)
	abline(h=seq(0,180,10),lty=2,col="gray")
}

